/*  RF.C  - Manages the CYRF6936 for communication
 *
 * You need the Technical Reference Manual for Cypress WirelessUSB LP devices to understand this
 *
 */

#include <xc.h>
#include "AES.h"
#include "spi.h"
#include "rf.h"
#include "main.h"
#include "sys_functions.h"
#include "nalig_protocol.h"

extern volatile unsigned char IRQ_flag;
extern volatile unsigned char timer2_flag;
extern volatile unsigned char timer5_flag;
extern volatile int buttonFlag;
extern unsigned int delayCount;

extern sys_config myConfig;
extern unsigned char deviceIsKeyfob;

unsigned char mfg_id[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
unsigned char CRC[2] = {0x00, 0x00}; //set to zero so we can figure it out...
unsigned char freq_1;
unsigned char currentPowerLevel=RF_MIN_PWR+1;
unsigned char crap[16];
unsigned char lastRSSI;
unsigned char lastChannel = 255;

// *************************************
// Change this.  CYRF6936 datahsheets has example SOP and PN sequences.
// *************************************
unsigned char soppn_sequence[24]= {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17};

void CYRF6936_init()
{
    spiSendCommand(0x9D, 0x01); //RESET!!!
    rest(TIMEOUT_1MS);
    //spiSendCommand(0xA8, 0x02); //CLK_EN_ADR, writes 2 because this is a streaming application
    spiSendCommand(0xB2, 0x3C); //AUTO_CAL_TIME_ADR...MUST write 0x3C to this register during initalization. So we will.
    spiSendCommand(0xB5, 0x14); //AUTO_CAL_OFFSET_ADR must write 0x14 to this during initalization...
    spiSendCommand(0x8D, 0x44); //sets IRQ polarity to be active high, PACTL is GPIO
    spiSendCommand(0x86, 0x48); //RX_CFG_ADR, LNA enabled, fast turn-around enabled
    spiSendCommand(0x9B, 0x55); //TX_OFFSET_LSB_ADR, increases TX freq by 1MHz. used to avoid swithing between freqs for TX/RX.
    spiSendCommand(0x9C, 0x05); //TX_OFFSET_MSB_ADR, typ. loaded with 0x05 (which it is)

    CRC[0] = 0xF5;
    CRC[1] = 0xDC;

    spiSendCommand2(0xD5, CRC[0], CRC[1]);  // CRC seed
    //spiSendCommand(0x95, CRC[0]);
    //spiSendCommand(0x96, CRC[1]);
    loadSOPPN();  // load the SOP and PN

    spiSendCommand(0x83, 0x2F); // TX_CFG_ADR,64 chip codes, 8DR mode, +4dBm
    spiSendCommand(0x92, 0x04); // Data64_THOLD_ADR, 64chip dataPN code correlator. typical = 0x07 (0A?)

    spiSendCommand(0xA4, 0x04); // framing preamble
    spiSendCommand(0xA4, 0x33); // framing preamble
    spiSendCommand(0xA4, 0x33); // framing preamble

    spiSendCommand(0x90, 0xE4); // FRAMING_CFG_ADR, SOP enable, SOP useses 64chips, Packet length EN, SOP correlator thesh is 0x0E
    spiSendCommand(0x9F, 0x00); // TX_OVERRIDE_ADR, disable all TX overrides
    spiSendCommand(0x9E, 0x02); // RX_OVERRIDE_ADDR, Accept bad CRC

    spiSendCommand(0x8F, 0xA7); // XACT_CFG_ADR, force state to Idle, Auto-ack enabled, auto-ack timeout to max (15x)
    spiSendCommand(0x8C, 0xC0); // XTAL_CTRL_ADR, set XOUT to GPIO (for power saving)
    spiSendCommand(0x8E, 0x00); // GPIO_CTRL_ADR, set all GPIO to input (for power saving)
    spiSendCommand(0x8B, 0x00); // PWR_CTRL_ADR: disable PMU
}


unsigned char CYRF6936_mfgid()
{
    
    // reads the device MFG ID
    // returns 1 if valid, 0 if invalid
    spiSendCommand(0xA5, 0xFF);  // unlock register
    spiReceiveBurstData(0x25, mfg_id, 6);  // read MFG ID
    spiSendCommand(0xA5, 0x00);  // lock register

    // count number of bits in 2 bytes of the mfg id
    unsigned int testInt = mfg_id[1] + (mfg_id[2] << 8);
    unsigned char bitCount;
    
    for (bitCount = 0; testInt; testInt >>= 1)
    {
      bitCount += testInt & 1;
    }

    //if (mfg_id[0] + mfg_id[1] + mfg_id[2] + mfg_id[3] + mfg_id[4] + mfg_id[5])
    if (bitCount > 2)
    {
        return 1;   // valid
    } else {
        return 0;   // invalid
    }
}

void CYRF6936_sleep()
{
    //spiSendCommand(0x8B, 0x00); // PWR_CTRL_ADR: disable PMU
    spiSendCommand(0x8F, 0x20); // enter sleep mode
    SPI_CS = 0;
    PWR_ON = 0;
}

void CYRF6936_wakeup()
{
    PWR_ON = 1;
    SPI_CS = 1;
    //spiSendCommand(0x8F, 0xA7); // XACT_CFG_ADR, set state to Idle, Auto-ack enabled, auto-ack timeout to max (15x)
    return;
    //spiSendCommand(0x8B, 0x80); // PWR_CTRL_ADR: enable PMU
    waitForSynth();
}

void waitForSynth()
{
    // uses an interrupt to wait for the synthesizer to settle.  max wait time is 5ms
    spiSendCommand(0x8C, 0xE0); // XTAL_CTRL_ADR: crystal stable interrupt enable
    if (!IRQ)  { // if the synthesizer is already settled, the IRQ line will go high immediately and we don't need to wait
        sleepUntilIRQ(TIMEOUT_5MS); // wait for crystal to stabilize (5ms max)
    }
    spiSendCommand(0x8C, 0xC0); // XTAL_CTRL_ADR: clear crystal stable interrupt
}


void loadSOPPN()
{
    // load the SOP and PN code
    unsigned char i;

    for (i=0; i<SOP_LENGTH; i++) // load the SOP
    {
        spiSendCommand(0xA2, soppn_sequence[i]);
    }

    for (i=SOP_LENGTH; i<SOP_LENGTH+PN_LENGTH; i++)  // load the PN
    {
        spiSendCommand(0xA3, soppn_sequence[i]);
    }
}

int TX (unsigned char channel, unsigned char *payload, unsigned char payloadSize)
{
    // transmit payload data on a specific frequency
    // returns power level sent at if successful, -1 if failed
    unsigned char tx_irq_stat=0;
    unsigned char nextByte=0, bytesRemaining=payloadSize, bytesToSend=0;

    if (channel != lastChannel)
    {
        spiSendCommand(0x80, channel);  // set channel (2400 + channel MHz)
        lastChannel = channel;
    }

    //spiSendCommand(0x8F, 0x03 | ACKflag); // XACT_CFG_ADR, end state will be sleep, Auto-ack enabled (if in flag), auto-ack timeout to max (15x)
    //spiSendCommand(0x8F, 0x04 ); // XACT_CFG_ADR, end state will be idle,
    //writeReg(XACT_CFG_ADR, 0x28);   // set mode to TX synth
    writeReg(XACT_CFG_ADR, 0x30);   // set mode to RX for RSSI reading

    spiSendCommand(0x81, payloadSize);  // TX_LENGTH_ADR, set packet length
    spiSendCommand(0x83, 0x28 | currentPowerLevel); // adjust transmit power

    while (!timer5_flag) Idle();    // wait for timer 5, we must wait at least 1ms between transmissions

    waitForSynth(); // wait for the synthesizer to settle before taking RSSI reading
    readReg(RSSI_ADR);  // dummy register read
    while ((readReg(RSSI_ADR) & 0x1F) > RSSI_THRESH) // is someone else transmitting?
    {
        //rest(TIMEOUT_200US);    // wait 200us before another RSSI reading
        SMALL_RND_DELAY;
        //rssi_flag = 1;
    }
    //if (rssi_flag) { rest(TIMEOUT_500US); rssi_flag = 0; }

    //GPIO1 = 1;
    if (myConfig.hasRFE) writeReg(GPIO_CTRL_ADR, 0xA0); //set PACTL and XOUT high (puts CC2591 in TX mode)


    spiSendCommand(0x82, 0xD7);     // TX_CTRL_ADR: TX GO, TX CLR, TXB8, TXBERR, TXC, TXE
    bytesRemaining = payloadSize;   // number of bytes to transmit

    // we will put 8 bytes into the TX FIFO at time, and wait for the TXB8 interrupt to tell us when to add more data
    while (bytesRemaining)
    {
        bytesToSend = (bytesRemaining > 8) ? 8 : bytesRemaining;     // send 8 bytes this time (or less)
        spiSendBurstCommand(0xA0, &payload[nextByte], bytesToSend);  // add data to TX FIFO
        nextByte += bytesToSend;
        bytesRemaining -= bytesToSend;

        // if we don't have any more data to add, disable the TXB8 interrupt and wait for a TXC (or error)
        if (!bytesRemaining) spiSendCommand(0x82, 0x07);  // TX_CTRL_ADR: TXBERR, TXC, TXE

        if (sleepUntilIRQ(TIMEOUT_10MS)) { // we should interrupt when the TX buffer is empty (or finished TX/error)
            tx_irq_stat =  spiSendCommand(0x04, 0x00);  // read TX_IRQ_STATUS_ADR
            if (tx_irq_stat & 0x07) break;  // if TXBERR, TXC, or TXE, we are done
        }
        else
        {
            if (myConfig.hasRFE)  writeReg(GPIO_CTRL_ADR, 0x00); //disable XOUT and PACTL (thus, disable the CC2591)
            return RF_TX_FAIL;  // TX timed out.  Serious radio config issues, should not get here
        }
    }

    //GPIO1 = 0;
    TMR5 = 0;           // reset timer 5.  at least 1ms must pass before we will be able to transmit again
    timer5_flag = 0;
    T5CONbits.TON = 1;

    if (myConfig.hasRFE)  writeReg(GPIO_CTRL_ADR, 0x00); //disable XOUT and PACTL (thus, disable the CC2591)


    if ((tx_irq_stat & 0x05) )  //if TXE or TXBERR
    {
            return RF_TX_FAIL;  // probably won't get here
    }
    return currentPowerLevel;
}

int RX (unsigned char channel, unsigned char *payload, unsigned int timeout)
{
    // try to receive data for up to the timeout
    // timeout is based on timer2, use the TIMEOUT macros (default: TIMEOUT_20MS)
    // returns the number of bytes receieved (0 if nothing, duh)
    unsigned char rx_irq_stat, rx_stat, rx_size, rx_expected_size;

    unsigned char bytesReceived=0, bytesRemaining=0;
    
    if (channel != lastChannel)
    {
        spiSendCommand(0x80, channel);  // set channel (2400 + channel MHz)
        lastChannel = channel;
    }
    //spiSendCommand(0x8F, 0x03 | ACKflag); // XACT_CFG_ADR, end state will be sleep, Auto-ack enabled, auto-ack timeout to max (15x)
    if (myConfig.hasRFE) writeReg(GPIO_CTRL_ADR, 0x80); //set PACTL low and XOUT high (puts CC2591 in RX mode)
    spiSendCommand(0x8F, 0x04); // XACT_CFG_ADR, end state will be idle,
    //GPIO2 = 1;
    //spiSendCommand(0x85, 0x83); // start RX, configure RX interrupts
    spiSendCommand(0x85, 0x93); // start RX, configure RX interrupts
    lastRSSI = spiSendCommand(0x13, 0x01); //read RSSI

    // wait for the timeout period to receive something
    waitForMoreData:
    if (sleepUntilIRQ(timeout))  // we might have gotten some data
    {
        rx_irq_stat = spiSendCommand(0x07, 0x00);   // read RX_IRQ_STAT_ADR
        if (rx_irq_stat & 0x07) goto finishRX;  // if RXE, RXC, or RXBERR, we are done

        if (rx_irq_stat & 0x10) //RXB8 
        {
            spiReceiveBurstData(0x21, &payload[bytesReceived], 8);
            bytesReceived += 8;
            timeout = TIMEOUT_2MS;
            goto waitForMoreData;
        }

        finishRX:
        rx_expected_size = spiSendCommand(0x0A, 0x00);      //RX_LENGTH: number of bytes expected
        rx_size = spiSendCommand(0x09, 0x00);               //RX_COUNT_ADR: number of bytes receieved

        if (rx_irq_stat & 0x01) // is there an error?
        {
            rx_stat = spiSendCommand(0x08, 0x00);   // we must read RX_STATUS_ADR to clear the error
        }

        if (bytesReceived < rx_size) {   // still have some data in the buffer?
            bytesRemaining = rx_size - bytesReceived;
            spiReceiveBurstData(0x21, &payload[bytesReceived], bytesRemaining);
            bytesReceived += bytesRemaining;
        }

        //GPIO2 = 0;
        if (bytesReceived == rx_expected_size && (rx_irq_stat & 0x02) && !(rx_irq_stat & 0x04))  // valid data
        {
            writeReg(XACT_CFG_ADR, 0x2C); // go to RX synth mode for fast turnaround (should not have to do this!!!!!!!!!!!!!)
            if (myConfig.hasRFE)  writeReg(GPIO_CTRL_ADR, 0x00); //disable XOUT and PACTL (thus, disable the CC2591)
            return bytesReceived;
        }

    } else {    // did not receive data in time allowed, turn off the radio's RX
        //GPIO2 = 0;
        writeReg(RX_ABORT_ADR, 0x20);       // abort enable
        readReg(RSSI_ADR);
        //while (readReg(RSSeadReg(RSSI_ADR);I_ADR) & 0x80 );   // RSSI_ADR make sure we aren't getting an SOP

        //writeReg(XACT_CFG_ADR, 0x20); // go to sleep
        writeReg(XACT_CFG_ADR, 0x24); // go idle
        writeReg(RX_ABORT_ADR, 0x00); // abort disable
    }

    if (readReg(RX_IRQ_STATUS_ADR) & 0x08)  // uh, is there still crap in the buffer?
    {
        spiReceiveBurstData(0x21, crap, 16); // dummy read from RX buffer
    }

    if (myConfig.hasRFE)  writeReg(GPIO_CTRL_ADR, 0x00); //disable XOUT and PACTL (thus, disable the CC2591)
    return 0;
}

unsigned char sleepUntilIRQ(unsigned int timeout)
{
    // goes to sleep, waiting for IRQ pin interrupt (low to high)
    // returns 1 if the interrupt happened, 0 if timeout

    if (IRQ)  return 1;   // if the IRQ pin is already high, we already have an interrupt

    CNEN2bits.CN21IE = 1; // enable input change interrupt for the IRQ pin

    T2CONbits.TON = 1;          // enable timer 2 for a timeout
    TMR2 = timeout;
    IRQ_flag = 0;
    timer2_flag = 0;

    while (!IRQ_flag && !timer2_flag)
        Idle();    // go to sleep, wait for timer or IRQ interrupts ONLY

    T2CONbits.TON = 0;          // disable timer 2

    CNEN2bits.CN21IE = 0; // disable input change interrupt for the IRQ pin

    return IRQ_flag; 
}
